package com.perforce.hws.swagger;

import io.swagger.models.Swagger;
import io.swagger.parser.SwaggerParser;
import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroupFile;

import com.perforce.hws.swagger.models.Definition;
import com.perforce.hws.swagger.models.Route;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Usage: java ... GenerateJavaScriptClient source_yaml output_dir
 */
public class GenerateJavaScriptClient implements HWSSwaggerMethods {

    public static void main(String args[]) {
        try {
            new GenerateJavaScriptClient(args[1], args[0]).run();
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    private String sourceYaml;
    private String outputDir;

    public GenerateJavaScriptClient(String outputDir, String sourceYaml) {
        this.outputDir = outputDir;
        this.sourceYaml = sourceYaml;
    }

    public void run() throws Exception {
        if (!Files.exists(Paths.get(outputDir))) {
            Files.createDirectories(Paths.get(outputDir));
        }

        Swagger swagger = new SwaggerParser().read(sourceYaml);

        createApiClient(swagger);

        List<Definition> definitions = convertToDefinitions(swagger);
        createModels(definitions);

        List<Route> routes = convertToRoutes(swagger);
        createApi(false, routes);
        createApi(true, routes);
    }

    private void createApiClient(Swagger swagger) throws IOException {
        STGroupFile apiClientGroup = new STGroupFile("javascript/api_client.stg");

        ST template = apiClientGroup.getInstanceOf("api_client");

        template.add("basePath", swagger.getBasePath());

        String content = template.render();

        template.remove("basePath");

        Files.write(apiClientPath(), content.getBytes());
    }

    private void createModels(List<Definition> definitions) throws IOException {
        STGroupFile modelGroup = new STGroupFile("javascript/model.stg");

        ST template = modelGroup.getInstanceOf("model");

        for (Definition definition : definitions) {
            template.add("definition", definition);

            String content = template.render();

            template.remove("definition");

            Path modelPath = modelPath(definition);
            if (!Files.exists(modelPath.getParent())) {
                Files.createDirectories(modelPath.getParent());
            }
            Files.write(modelPath, content.getBytes());
        }

        ST exportTemplate = modelGroup.getInstanceOf("export_all");

        exportTemplate.add("definitions", definitions);

        String exportContent = exportTemplate.render();

        Files.write(modelsPath(), exportContent.getBytes());
    }

    private void createApi(boolean isAlpha, List<Route> routes) throws IOException {
        STGroupFile apiGroup = new STGroupFile("javascript/api.stg");

        ST apiTemplate = apiGroup.getInstanceOf("api");

        apiTemplate.add("apiClass", isAlpha ? "AlphaApi" : "DefaultApi");
        apiTemplate.add("routes",
                routes.stream().filter(r -> {
                    if (isAlpha) {
                        return r.isAlpha();
                    } else {
                        return !r.isAlpha();
                    }
                })
                .collect(Collectors.toList())
        );

        String content = apiTemplate.render();

        Files.write(apiPath(isAlpha), content.getBytes());
    }

    private Path apiPath(boolean isAlpha){
        return Paths.get(outputDir, "lib", isAlpha ? "alpha_api.js" : "default_api.js");
    }

    private Path apiClientPath() {
        return Paths.get(outputDir, "lib", "api_client.js");
    }

    private Path modelPath(Definition definition) {
        return Paths.get(outputDir, "lib", "models", definition.getJsFileName());
    }

    private Path modelsPath() {
        return Paths.get(outputDir, "lib", "models.js");
    }
}
